'use strict';

var _exports;

//---------------------------------------------------------------------
//
// This source file is part of the Meta:Magical project.
//
// See LICENCE for licence information.
// See CONTRIBUTORS for the list of contributors to the project.
//
//---------------------------------------------------------------------

// -- DEPENDENCIES ----------------------------------------------------
var generate = require('babel-generator').default;
var template = require('babel-template');
var traverse = require('babel-traverse').default;

var _require = require('babylon');

var parse = _require.parse;

// -- HELPERS ---------------------------------------------------------

/*~
 * A template for creating an assertion AST.
 */

var buildAssertion = template('\n  __metamagical_assert_equals(require(MODULE), ACTUAL, EXPECTED, MESSAGE)\n');

/*~
 * The deep equality function.
 */
buildAssertion[Symbol.for('@@meta:magical')] = {
  'name': 'buildAssertion',
  'documentation': 'A template for creating an assertion AST.\n '
};
function assertEquals(assert, actual, expected, message) {
  var assertionType = Symbol.for('@@meta:magical:assertion-type');
  var rest = Symbol.for('@@meta:magical:assertion-rest');
  var keys = Object.keys;

  function isSetoid(value) {
    return value && typeof value.equals === 'function';
  }

  function isRecord(value) {
    return value && value[assertionType] === 'record';
  }

  function check(predicate) {
    for (var _len = arguments.length, values = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      values[_key - 1] = arguments[_key];
    }

    return values.every(predicate);
  }

  function getSpecialArrayLength(array) {
    var rests = array.map(function (x, i) {
      return [x, i];
    }).filter(function (pair) {
      return pair[0] === rest;
    });
    if (rests.length > 1 || rests[0] && rests[0][1] !== array.length - 1) {
      assert.ok(false, message);
    }
    return (rests[0] || [null, array.length])[1];
  }

  function compareArrays(left, right) {
    var expectedLength = getSpecialArrayLength(right);
    assert.ok(Array.isArray(left), message);
    if (left.length < expectedLength) {
      assert.ok(false, message);
    }
    for (var i = 0; i < expectedLength; ++i) {
      compare(left[i], right[i]);
    }
  }

  function compareRecord(left, right) {
    var isPartial = right[rest];
    assert.ok(Object(left) === left, message);
    if (!isPartial) {
      assert.deepStrictEqual(keys(left).sort(), keys(right).sort(), message);
    }
    Object.keys(right).forEach(function (key) {
      if (!(key in left)) {
        assert.ok(false, message);
      }
      compare(left[key], right[key]);
    });
  }

  function compare(l, r) {
    return check(isSetoid, l, r) ? assert.ok(l.equals(r), message) : Array.isArray(r) ? compareArrays(l, r) : isRecord(r) ? compareRecord(l, r) : /* otherwise */assert.deepStrictEqual(l, r, message);
  }

  compare(actual, expected);
}

/*~
 * Returns the first item of an array.
 */
assertEquals[Symbol.for('@@meta:magical')] = {
  'name': 'assertEquals',
  'documentation': 'The deep equality function.\n '
};
function first(xs) {
  return xs[0];
}

/*~
 * Tests if a comment node in the AST is an assertion comment.
 */
first[Symbol.for('@@meta:magical')] = {
  'name': 'first',
  'documentation': 'Returns the first item of an array.\n '
};
function isAssertion(comment) {
  return comment.type === 'CommentLine' && /^\s*==>\s*.+$/.test(comment.value);
}

// -- IMPLEMENTATION --------------------------------------------------
/*~
 * A Babel transformer that converts assertion comments into runtime
 * assertions.
 *
 * Assertion comments are usually used in examples to show readers
 * what a particular expression evaluates to, but they don't get
 * checked when one runs the code. This plugin allows people to write
 * assertion comments after a particular expression. It then transforms
 * this assertion comment into a real runtime assertion.
 *
 *     2 * 2  // ==> 4
 *
 *
 * ## Options
 *
 * You can configure which assertion module is used by providing the
 * `module` option to Babel. By default this transform uses Node's
 * built-in `assert` module.
 */
isAssertion[Symbol.for('@@meta:magical')] = {
  'name': 'isAssertion',
  'documentation': 'Tests if a comment node in the AST is an assertion comment.\n '
};
module.exports = (_exports = function _exports(_ref) {
  var t = _ref.types;

  var assertEqualsFD = parse(assertEquals.toString()).program.body[0];
  var assertEqualsAST = t.functionExpression(t.identifier('metamagical_assert_equals'), assertEqualsFD.params, assertEqualsFD.body);

  function makeSymbol(name) {
    return t.callExpression(t.memberExpression(t.identifier('Symbol'), t.identifier('for')), [t.stringLiteral(name)]);
  }

  function makeRest() {
    return makeSymbol('@@meta:magical:assertion-rest');
  }

  function makeAssertionType() {
    return makeSymbol('@@meta:magical:assertion-type');
  }

  function transformSpread(ast) {
    traverse(ast, {
      enter: function enter(path) {
        switch (path.node.type) {
          case 'SpreadElement':
            if (path.node.argument.name === "_") {
              path.replaceWith(makeRest());
            }
            break;

          case 'ObjectExpression':
            path.node.properties.push(t.objectProperty(makeAssertionType(), t.stringLiteral('record'), true));
            break;

          case 'SpreadProperty':
            if (path.node.argument.name === "_") {
              path.replaceWith(t.objectProperty(makeRest(), t.booleanLiteral(true), true));
            }
            break;
        }
      }
    });
    return ast;
  }

  function parseAssertion(source) {
    var data = source.replace(/^\s*==>\s*/, '').trim();

    try {
      var ast = parse('(' + data + ')', { plugins: ['objectRestSpread'] });
      transformSpread(ast);
      ast = ast.program.body;
      if (ast.length !== 1 && ast[0].type !== 'ExpressionStatement') {
        throw new TypeError('Expected a single expression in ' + data);
      }

      return {
        expression: ast[0].expression,
        source: data
      };
    } catch (e) {
      console.log('Error parsing ' + source + ': ', e);
      return null;
    }
  }

  function getTrailingAssertion(node) {
    var comment = first(node.trailingComments || []);

    return comment && isAssertion(comment) ? parseAssertion(comment.value) : /* otherwise */null;
  }

  return {
    visitor: {
      ExpressionStatement: function ExpressionStatement(path, state) {
        var assertModule = state.opts.module || 'assert';
        var assertion = getTrailingAssertion(path.node);

        if (assertion) {
          if (!path.hub.file.scope.hasBinding('__metamagical_assert_equals')) {
            path.hub.file.scope.push({
              id: t.identifier('__metamagical_assert_equals'),
              init: assertEqualsAST
            });
          }

          path.node.expression = buildAssertion({
            MODULE: t.stringLiteral(assertModule),
            ACTUAL: path.node.expression,
            EXPECTED: assertion.expression,
            MESSAGE: t.stringLiteral(generate(path.node.expression).code + ' ==> ' + assertion.source)
          });
        }
      }
    }
  };
}, _exports[Symbol.for('@@meta:magical')] = {
  'name': 'exports',
  'documentation': 'A Babel transformer that converts assertion comments into runtime\nassertions.\nAssertion comments are usually used in examples to show readers\nwhat a particular expression evaluates to, but they don\'t get\nchecked when one runs the code. This plugin allows people to write\nassertion comments after a particular expression. It then transforms\nthis assertion comment into a real runtime assertion.\n    2 * 2  // ==> 4\n## Options\nYou can configure which assertion module is used by providing the\n`module` option to Babel. By default this transform uses Node\'s\nbuilt-in `assert` module.\n '
}, _exports);